# !/usr/bin/python3
"""
Contains necessary objects for creating a Directed Acyclic Graph (DAG) object to represent Open Biomedical Ontologies
(OBO).
"""
import re


class OboGraph(object):

    """A pythonic graph of a generic Open Biomedical Ontology (OBO) directed
    acyclic graph (DAG).
    """

    def __init__(self, namespace_filter=None, allowed_relationships=None):
        """`OboGraph` initializer. Leave `namespace_filter` and `allowed_relationship` as :py:obj:`None` to create the
        entire ontology graph. Otherwise, provide filters to limit what information is pulled into the graph.

        :param str namespace_filter: Specify the namespace of a sub-ontology namespace, if one is available for the ontology.
        :param list allowed_relationships: Specify a list of relationships to utilize in the graph, other relationships will be ignored.
        """
        self.namespace_filter = namespace_filter
        self.allowed_relationships = allowed_relationships
        self.word_split = re.compile(r"[\w\'\-]+")
        self.node_list = list()
        self.edge_list = list()
        self.id_index = dict()
        self.vocab_index = dict()
        self.relationship_index = dict()
        self.used_relationship_set = set()
        self.relationship_count = dict()
        self.root_nodes = list()
        self._orphans = None
        self._leaves = None
        self._modified = True

        if self.allowed_relationships:
            if 'is_a' not in self.allowed_relationships:
                print("WARNING: 'is_a' is a required relationship type within OBO ontologies.\nAdding 'is_a' to the allowed_relationships list.")
                self.allowed_relationships.append('is_a')

    @property
    def orphans(self):
        """:py:obj:`property` defining a set of nodes in the graph which have no parents. When the graph is modified,
        calls :func:`_update_graph` to repopulate the sets of orphan and leaf nodes.

        :return: Set of 'orphan' :class:`gocats.dag.AbstractNode` objects.
        :rtype: :py:class:`set`
        """
        if self._modified:
            self._update_graph()
        return self._orphans

    @property
    def leaves(self):
        """:py:obj:`property` defining a set of nodes in the graph which have no children. When the graph is modified,
        calls :func:`_update_graph` to repopulate the sets of orphan and leaf nodes.

        :return: Set of 'leaf' :class:`gocats.dag.AbstractNode` objects.
        :rtype: :py:class:`set`
        """
        if self._modified:
            self._update_graph()
        return self._leaves

    def valid_node(self, node):
        """Defines condition of a valid node. Node is valid if it is not obsolete and is contained within the given
        ontology namespace constraint.

        :param node: A :class:`gocats.dag.AbstractNode` object
        :return: True if node is valid, False otherwise
        :rtype: :py:obj:`True` or :py:obj:`False`
        """
        if not node.obsolete and (not self.namespace_filter or node.namespace == self.namespace_filter):
            return True
        return False

    def valid_edge(self, edge):
        """Defines condition of a valid edge. Edge is valid if it is within the list of allowed edges and connects two
        nodes that are both contained in the graph in question.

        :param edge: A :class:`gocats.dag.AbstractEdge` object
        :return: True if node is valid, False otherwise
        :rtype: :py:obj:`True` or :py:obj:`False`
        """
        if (edge.parent_node.id in self.id_index and edge.child_node.id in self.id_index) and (not self.allowed_relationships or edge.relationship_id in self.allowed_relationships):
            return True
        return False

    def _update_graph(self):
        """Repopulates graph orphans and leaves sets.

        :return: None
        :rtype: :py:obj:`None`
        """
        self._orphans = set([node for node in self.node_list if not node.obsolete and not node.parent_node_set and node not in self.root_nodes])
        self._leaves = set([node for node in self.node_list if not node.obsolete and not node.child_node_set and node.parent_node_set])
        self._modified = False

    def add_node(self, node):
        """Adds a node object to the graph, adds an object pointer to the vocabulary index to reference nodes to every
        word in the node name and definition. Sets modification state to :py:obj:`True`.

        :param node: A :class:`gocats.dag.AbstractNode` object.
        :return: None
        :rtype: :py:obj:`None`
        """
        self.node_list.append(node)
        self.id_index[node.id] = node
        for word in re.findall(r"[\w\'\-]+", node.name + " " + node.definition):
            try:
                self.vocab_index[word].add(node)
            except KeyError:
                self.vocab_index[word] = set([node])  # Don't replace with set literal
        self._modified = True

    def remove_node(self, node):
        """Removes a node from the graph and deletes node references from all entries in the vocabulary index. Sets
        modification state to :py:obj:`True`.

        :param node: A :class:`gocats.dag.AbstractNode` object.
        :return: None
        :rtype: :py:obj:`None`
        """
        if node not in self.node_list:
            pass
        else:
            for graph_node in self.node_list:
                if node in graph_node.parent_node_set:
                    graph_node.parent_node_set.remove(node)
                elif node in graph_node.child_node_set:
                    graph_node.child_node_set.remove(node)
                for edge in graph_node.edges:
                    if node is edge.parent_node or node is edge.child_node:
                        graph_node.edges.remove(edge)
            for word in re.findall(r"[\w\'\-]+", node.name + " " + node.definition):
                try:
                    self.vocab_index[word].remove(node)
                except KeyError:
                    pass
                else:
                    if not self.vocab_index[word]:
                        del self.vocab_index[word]
            del self.id_index[node.id]
            self.node_list.remove(node)
        self._modified = True

    def add_edge(self, edge):
        """Adds an edge object to the graph, and counts the edge relationship type. Sets modification state to
        :py:obj:`True`.

        :param edge: A :class:`gocats.dag.AbstractEdge` object.
        :return: None
        :rtype: :py:obj:`None`
        """
        self.edge_list.append(edge)
        try:
            self.relationship_count[edge.relationship_id] += 1
        except KeyError:
            self.relationship_count[edge.relationship_id] = 1
        self._modified = True

    def remove_edge(self, edge):
        """Removes an edge object from the graph, and removes references to that edge from the node objects involved.
        Sets modification state to :py:obj:`True`.

        :param edge: A :class:`gocats.dag.AbstractEdge` object.
        :return: None
        :rtype: :py:obj:`None`
        """
        self.id_index[edge.parent_id].remove_edge(edge)
        self.id_index[edge.child_id].remove_edge(edge)
        self.edge_list.remove(edge)
        self._modified = True

    def add_relationship(self, relationship):
        """Adds a :class:`gocats.dag.AbstractRelationship` object to the graph's relationship index, referenced by
        that relationships ID. Sets modification state to :py:obj:`True`.

        :param relationship: A :class:`gocats.dag.AbstractRelationship` object.
        :return: None
        :rtype: :py:obj:`None`
        """
        self.relationship_index[relationship.id] = relationship
        self._modified = True

    def instantiate_valid_edges(self):
        """Add all edge references to their respective nodes and vice versa if both nodes of the edge are in the graph.
        This is carried out by :func:`AbstractEdge.connect_nodes`. Also adds :class:`gocats.dag.AbstractRelationship`
        object reference to each edge. If both nodes are not in the graph, the edge is deleted from the graph. Sets
        modification state to :py:obj:`True`.

        :return: None
        :rtype: :py:obj:`None`
        """
        del_edges = set()
        for edge in self.edge_list:
            if edge.node_pair_id[0] in self.id_index.keys() and edge.node_pair_id[1] in self.id_index.keys():
                edge.relationship = self.relationship_index[edge.relationship_id]
                edge.connect_nodes((self.id_index[edge.node_pair_id[0]], self.id_index[edge.node_pair_id[1]]), self.allowed_relationships)
            else:
                del_edges.add(edge)
        for edge in del_edges:
            self.edge_list.remove(edge)
        self._modified = True

    def node_depth(self, sample_node):
        """Returns an integer representing how many nodes are between the given node and the root node of the graph
        (depth level).

        :param sample_node: A :class:`gocats.dag.AbstractNode` object.
        :return: Depth level.
        :rtype: :py:obj:`int`
        """
        if sample_node in self.root_nodes:
            return 0
        depth = 1
        root_node_set = set(self.root_nodes)
        parent_set = sample_node.parent_node_set
        while parent_set:
            if parent_set & root_node_set:  # There is an intersection between the parent set and the root_node_set
                break
            depth += 1
            parent_set = set().union(*[parent.parent_node_set for parent in parent_set])
        return depth

    def filter_nodes(self, search_string_list):
        """Returns a list of node objects that contain vocabulary matching the keywords provided in the search string
        list. Nodes are selected by searching through the vocablary index.

        :param search_string_list: A :py:obj:`list` of search strings provided in the keyword_file provided to :func:`gocats.gocats.create_subgraphs`.
        :return: A list of :class:`gocats.dag.AbstractNode` objects.
        :rtype: :py:obj:`list`
        """
        search_string_list_words = [re.findall(self.word_split, word) for word in search_string_list]
        search_string_word_set = set([word for sublist in search_string_list_words for word in sublist])
        filtered_nodes = set.union(*[node_set for node_set in [self.vocab_index[word] for word in search_string_word_set if word in self.vocab_index]])
        if self.namespace_filter:
            filtered_nodes = [node for node in filtered_nodes if node.namespace == self.namespace_filter]
        return filtered_nodes

    def filter_edges(self, filtered_nodes):
        """Returns a list of edges in the graph that connect the nodes provided in the filtered nodes list.

        :param filtered_nodes: List of filtered nodes provided by :func:`filter_nodes`.
        :return: A list of :class:`gocats.dag.AbstractEdge` objects.
        :rtype: :py:obj:`list`
        """
        filtered_edges = [edge for edge in self.edge_list if edge.parent_node in filtered_nodes and edge.child_node in filtered_nodes]
        if self.allowed_relationships:
            filtered_edges = [edge for edge in filtered_edges if edge.relationship_id in self.allowed_relationships]
        return filtered_edges

    def nodes_between(self, start_node, end_node):
        """Returns a set of nodes that occur along all paths between the start node and the end node. If no paths exist,
        an empty set is returned.

        :param start_node: :class:`gocats.dag.AbstractNode` object to start the paths.
        :param end_node: :class:`gocats.dag.AbstractNode` object to end the paths.
        :return: A set of :class:`gocats.dag.AbstractNode` objects if there is at least one path between the parameters, an empty set otherwise.
        :rtype: :py:obj:`set`
        """
        if start_node.ancestors and end_node.descendants:
            return start_node.ancestors.intersection(end_node.descendants)
        else:
            return set()


class AbstractNode(object):

    """A node containing all basic properties of an OBO node. The parsing object, :class:`gocats.ontologyparser.OboParser`
    currently has direct access to data members (id, name, definition, namespace, edges, and obsolete) so that
    information from the database file can be added to the object.
    """

    def __init__(self):
        """`AbstractNode` initializer
        """
        self.id = str()
        self.name = str()
        self.definition = str()
        self.namespace = str()
        self.edges = set()
        self.parent_node_set = set()
        self.child_node_set = set()
        self.obsolete = False
        self._modified = True
        self._descendants = None
        self._ancestors = None
        # Will add new sets for equivalence, actor/actee, ordinal, etc

    @property
    def descendants(self):
        """:py:obj:`property` defining a set of nodes in the graph that are recursively reverse of a node with a
        scoping-type relationship. When the node is modified, calls :func:`gocats.dag.AbstractNode._update_node` to
        repopulate the sets of descendants and ancestors. This represents a "lazy" evaluation of node descendants.

        :return: Set of :class:`gocats.dag.AbstractNode` objects
        :rtype: :py:class:`set`
        """
        if self._modified:
            self._update_node()
        return self._descendants

    @property
    def ancestors(self):
        """:py:obj:`property` defining a set of nodes in the graph that are recursively forward of a node with a
        scoping-type relationship. When the node is modified, calls :func:`gocats.dag.AbstractNode._update_node` to
        repopulate the sets of descendants and ancestors. This represents a "lazy" evaluation of node ancestors.

        :return: Set of :class:`gocats.dag.AbstractNode` objects
        :rtype: :py:class:`set`
        """
        if self._modified:
            self._update_node()
        return self._ancestors

    def _update_node(self):
        """Repopulates ancestor and descendant sets for a node. Sets modification state to :py:obj:`True`.

        :return: None
        :rtype: :py:obj:`None`
        """
        self._update_descendants()
        self._update_ancestors()
        self._modified = False

    def add_edge(self, edge, allowed_relationships):
        """Adds a given :class:`gocats.dag.AbstractEdge` to a each :class:`gocats.dag.AbstractNode` objects that the
        edge connects. If there is a filter for the types of relationships allowed, edges with non-allowed relationship
        types are not processed. Sets modification state to :py:obj:`True`.

        :return: None
        :rtype: :py:obj:`None`
        """
        # TODO: Need to capture non-parent/child relationship types, such as actor/actee and equivalence
        # FIXME: Should we add edges that represent non-allowed relationships?
        self.edges.add(edge)
        if not allowed_relationships:
            if edge.child_id == self.id:
                self.parent_node_set.add(edge.parent_node)
            elif edge.parent_id == self.id:
                self.child_node_set.add(edge.child_node)
        else:
            if edge.child_id == self.id and edge.relationship_id in allowed_relationships:
                self.parent_node_set.add(edge.parent_node)
            elif edge.parent_id == self.id and edge.relationship_id in allowed_relationships:
                self.child_node_set.add(edge.child_node)
        self._modified = True

    def remove_edge(self, edge):
        """Removes a given :class:`gocats.dag.AbstractEdge` the :class:`gocats.dag.AbstractNode` object. Also removes
        parent or child node references that the edge referenced. Sets modification state to :py:obj:`True`.

        :return: None
        :rtype: :py:obj:`None`
        """
        if edge.child_id == self.id:
            self.parent_node_set.remove(edge.parent_node)
        elif edge.parent_id == self.id:
            self.child_node_set.remove(edge.child_node)
        self.edges.remove(edge)
        self._modified = True

    def _update_descendants(self):
        """Used for the lazy evaluation of graph descendants of the current :class:`gocats.dag.AbstractNode` object.
        Creates internal :py:obj:`set` variable, descendant_set. Iterates through node children until the bottom of the
        graph is reached. The descendant_set is a set of all nodes across all paths encountered from the current node.

        :return: None
        :rtype: :py:obj:`None`
        """
        descendant_set = set()
        children = list(self.child_node_set)
        while len(children) > 0:
            child = children[0]
            descendant_set.add(child)
            if not child._modified:
                descendant_set.update(child._descendants)
            else:
                children.extend([new_child for new_child in child.child_node_set if new_child not in descendant_set and new_child not in children])
            children.remove(child)
        self._descendants = descendant_set

    def _update_ancestors(self):
        """Used for the lazy evaluation of graph ancestors of the current :class:`gocats.dag.AbstractNode` object.
        Creates internal :py:obj:`set` variable, ancestors_set. Iterates through node parents until the top of the graph
        is reached. The ancestors_set is a set of all nodes across all paths encountered from the current node.

        :return: None
        :rtype: :py:obj:`None`
        """
        ancestors_set = set()
        parents = list(self.parent_node_set)
        while len(parents) > 0:
            parent = parents[0]
            ancestors_set.add(parent)
            if not parent._modified:
                ancestors_set.update(parent._ancestors)
            else:
                parents.extend([new_parent for new_parent in parent.parent_node_set if new_parent not in ancestors_set and new_parent not in parents])
            parents.remove(parent)
        self._ancestors = ancestors_set


class AbstractEdge(object):

    """An OBO edge which links two ontology term nodes and contains a relationship type describing now the two nodes are
     related.
    """

    def __init__(self, node1_id, node2_id, relationship_id, node_pair=None):
        """`AbstractEdge` initializer. Node pair refers to a :py:obj:`tuple` of :class:`gocats.dag.AbstractNode` objects that are
        connected by the edge. Defaults to :py:obj:`None` and is later populated.

        :param str node1_id: The ID of the first term referenced from the ontology file's relationship line.
        :param str node2_id: The ID of the second term referenced from the ontology file's relationship line.
        :param str relationship_id: The ID of the relationship in the ontology file's relationship line.
        :param tuple node_pair: Default-:py:obj:`None`, provide a :py:obj:`tuple` containing two :class:`gocats.dag.AbstractNode` objects if they are already created and able to be referenced.
        """
        self.node_pair_id = (node1_id, node2_id)
        self.node_pair = node_pair
        self.relationship_id = relationship_id
        self.relationship = None

    @property
    def json_edge(self):
        """:py:obj:`property` which returns a tuple where position 0 is a unique string representation of the edge made by combining the ID of the reverse node and the id of the forward nodes and where position 1 is a list of two node IDs: the reverse and forward node.

        :return: :py:obj:`tuple` of a unique :class:`AbstractEdge` ID and a list of that edge object's reverse and forward node IDs, respectively. Returns an empty :py:obj:str at a position for which there are no forward or reverse nodes in the graph.
        :rtype: :py:obj:`tuple`
        """
        reverse_node_id = self.reverse_node.id
        forward_node_id = self.forward_node.id
        return (str(reverse_node_id+forward_node_id), [reverse_node_id, forward_node_id])

    @property
    def parent_id(self):
        """:py:obj:`property` defining the ID of the node forward of the current :class:`gocats.dag.AbstractEdge`
        object.

        :return: :py:obj:`str` ID of the forward node in the node_pair associated with the edge if the edge's relationship is assigned, :py:obj:`None` otherwise.
        :rtype: :py:obj:`str` or :py:obj:`None`
        """
        if self.relationship:
            return self.relationship.forward(self.node_pair_id)
        return None

    @property
    def child_id(self):
        """:py:obj:`property` defining the ID of the node reverse of the current :class:`gocats.dag.AbstractEdge` object.

        :return: :py:obj:`str` ID of the reverse node in the node_pair associated with the edge if the edge's relationship is assigned, :py:obj:`None` otherwise.
        :rtype: :py:obj:`str` or :py:obj:`None`
        """
        if self.relationship:
            return self.relationship.reverse(self.node_pair_id)
        return None

    @property
    def forward_node(self):
        """:py:obj:`property` defining the :class:`gocats.dag.AbstractNode` object forward of the current
        :class:`gocats.dag.AbstractEdge` object.

        :return: :class:`gocats.dag.AbstractNode` object of the forward node in the node_pair associated with the edge if the edge's relationship is assigned, the node_pair is assigned, and the type of relationship is instantiated by :class:`gocats.dag.DirectionalRelationship` :py:obj:`None` otherwise.
        :rtype: :class:`gocats.dag.AbstractNode` or :py:obj:`None`
        """
        if self.node_pair and self.relationship and type(self.relationship) is DirectionalRelationship:
            return self.relationship.forward(self.node_pair)
        return None

    @property
    def reverse_node(self):
        """:py:obj:`property` defining the :class:`gocats.dag.AbstractNode` object reverse of the current
        :class:`gocats.dag.AbstractEdge` object.

        :return: :class:`gocats.dag.AbstractNode` object of the reverse node in the node_pair associated with the edge if the edge's relationship is assigned, the node_pair is assigned, and the type of relationship is instantiated by :class:`gocats.dag.DirectionalRelationship` :py:obj:`None` otherwise.
        :rtype: :class:`gocats.dag.AbstractNode` or :py:obj:`None`
        """
        if self.node_pair and self.relationship and type(self.relationship) is DirectionalRelationship:
            return self.relationship.reverse(self.node_pair)
        return None

    @property
    def parent_node(self):
        """:py:obj:`property` defining the :class:`gocats.dag.AbstractNode` object forward of the current
        :class:`gocats.dag.AbstractEdge` object. This designation will be unique to scoping-type relationships, although
        this is **not yet specified**.

        :return: :class:`gocats.dag.AbstractNode` object of the forward node in the node_pair associated with the edge if the edge's relationship is assigned, the node_pair is assigned, and the type of relationship is instantiated by :class:`gocats.dag.DirectionalRelationship` :py:obj:`None` otherwise.
        :rtype: :class:`gocats.dag.AbstractNode` or :py:obj:`None`
        """
        if self.relationship:
            return self.relationship.forward(self.node_pair)
        return None

    @property
    def child_node(self):
        """:py:obj:`property` defining the :class:`gocats.dag.AbstractNode` object reverse of the current
        :class:`gocats.dag.AbstractEdge` object. This designation will be unique to scoping-type relationships, although
        this is **not yet specified**.

        :return: :class:`gocats.dag.AbstractNode` object of the reverse node in the node_pair associated with the edge if the edge's relationship is assigned, the node_pair is assigned, and the type of relationship is instantiated by :class:`gocats.dag.DirectionalRelationship` :py:obj:`None` otherwise.
        :rtype: :class:`gocats.dag.AbstractNode` or :py:obj:`None`
        """
        if self.relationship:
            return self.relationship.reverse(self.node_pair)
        return None

    # Will finish these later
    @property
    def actor_node(self):
        """**not yet implemented**

        :return: None
        :rtype: :py:obj:`None`
        """
        return

    @property
    def recipient_node(self):
        """**not yet implemented**

        :return: None
        :rtype: :py:obj:`None`
        """
        return

    @property
    def ordinal_prior_node(self):
        """**not yet implemented**

        :return: None
        :rtype: :py:obj:`None`
        """
        return

    @property
    def ordinal_post_node(self):
        """**not yet implemented**

        :return: None
        :rtype: :py:obj:`None`
        """
        return

    @property
    def other_node(self, node):
        """**not yet implemented**

        :return: None
        :rtype: :py:obj:`None`
        """
        return

    def connect_nodes(self, node_pair, allowed_relationships):
        """Adds the current edge object to the :class:`gocats.dag.AbstractNode` objects that are connected by the edge.
        Populates the node_pair with :class:`gocats.dag.AbstractNode` objects.

        :return: None
        :rtype: :py:obj:`None`
        """
        self.node_pair = node_pair
        node_pair[0].add_edge(self, allowed_relationships)
        node_pair[1].add_edge(self, allowed_relationships)


class AbstractRelationship(object):

    """A relationship as defined by a [typedef] stanza in an OBO ontology and augmented by GOcats to better interpret
    semantic correspondence.
    """

    def __init__(self):
        """`AbstractRelationship` initializer.
        """
        self.id = str()
        self.name = str()
        self.category = str()  # TODO: change category to correspondence_classes DO everywhere.


class DirectionalRelationship(AbstractRelationship):

    """A singly-directional relationship edge connecting two nodes in the graph. The two nodes are designated 'forward'
    and 'reverse.' The 'forward' node semantically succeeds the 'reverse' node in a way that depends on the context of
    the type of relationship describing the edge to which it is applied.
    """

    def __init__(self):
        """`DirectionalRelationship` initializer.
        """
        super().__init__()
        self.inverse_relationship_id = None
        self.inverse_relationship = None
        self.direction = 1  # Defaults as toward node2 (node2 is the 'forward' node)

    def forward(self, pair):
        """Returns the forward node in a node pair that semantically succeeds the other and is independent of the
        directionality of the edge. Default position is the second position [1].

        :param tuple pair: A pair of :class:`gocats.dag.AbstractNode` objects.
        :return: The forward :class:`gocats.dag.AbstractNode` object as determined by the pre-defined semantic directionality of the relationship.
        """
        return pair[self.direction]

    def reverse(self, pair):
        """Returns the reverse node in a node pair that semantically precedes the other and is independent of the
        directionality of the edge. Default position is the second position [1].

        :param tuple pair: A pair of :class:`gocats.dag.AbstractNode` objects.
        :return: The reverse :class:`gocats.dag.AbstractNode` object as determined by the pre-defined semantic directionality of the relationship.
        """
        return pair[(self.direction+1) % 2]


class NonDirectionalRelationship(AbstractRelationship):

    """A non-directional relationship whose edge directionality is either non-existent or semantically irrelevant.
    """
    def __init__(self):
        """`NonDirectionalRelationship` initializer.
        """
        return



# !/usr/bin/python3
"""
Functions for handling some file input and output and reformatting tasks in GOcats.
"""
import json
import jsonpickle
import sys
import os
import re
import csv
maxInt = sys.maxsize

while True:
    # decrease the maxInt value by factor 10
    # as long as the OverflowError occurs.

    try:
        csv.field_size_limit(maxInt)
        break
    except OverflowError:
        maxInt = int(maxInt/10)


# TODO: move to using JSON not JsonPickle and use sort_keys=True parameter to test outputs between runs
def json_save(obj, filename):
    """Takes a Python object, converts it into a JSON serializable object (if it is not already), and saves it to a file
    that is specified.

    :param obj: A Python :py:obj:`obj`.
    :param file_handle filename: A path to output the resulting JSON file.
    """
    if type(obj) == str or type(obj) == list:
        json_obj = obj
    elif type(obj) == dict:
        json_obj = dict()
        for key, value in obj.items():
            if type(value) == str or type(value) == list:
                new_value = value
            elif type(value) == set:
                new_value = [item for item in value]
            else:
                raise Exception("Data type is not supported!")
            json_obj[key] = new_value
    elif type(obj) == set:
        json_obj = [item for item in obj]
    else:
        raise Exception("Data type is not supported!")
    with open(filename+".json", 'w') as json_file:
        json_text = json.dumps(json_obj, sort_keys=True)
        json_file.write(json_text)


def jsonpickle_save(obj, filename):
    """Takes a Python object, converts it into a JsonPickle string, and writes it out to a file.

    :param obj: A Python :py:obj:`obj`
    :param file_handle filename:  A path to output the resulting JsonPickle file.
    """
    f = open(filename+".json_pickle", 'w')
    json_obj = jsonpickle.encode(obj, keys=True)  # Use_jsonpickle=True used to prevent jsonPickle from encoding dictkeys to strings.
    f.write(json_obj)
    f.close()


def jsonpickle_load(filename):
    """Takes a JsonPickle file and loads in the JsonPickle object into a Python object.

    :param file_handle filename: A path to a JsonPickle file.
    """
    f = open(filename)
    json_str = f.read()
    obj = jsonpickle.decode(json_str, keys=True)  # Use_jsonpickle=True used to prevent jsonPickle from encoding dictkeys to strings.
    return obj


def list_to_file(filename, data):
    """Makes a text document from a :py:obj:`list`  of data, with each line of the document being one item from the list
     and outputs the document into a file.

     :param file_handle filename: A path to the output file.
     :param data: A Python :py:obj:`list`.
     """
    with open(filename + ".txt", 'wt') as out_file:
        for line in data:
            out_file.write(str(line) + '\n')


# Functions for handling Gene Annotation Files
def write_out_gaf(data, filename):
    """Writes out an object representing a Gene Annotation File (GAF) to a file.

    :param list data: A :py:obj:`list` object representing a GAF. Each item in the list represents a row.
    :param file_handle filename: A path and name for the GAF.
    """
    with open(filename, 'w') as gaf_file:
        gafwriter = csv.writer(gaf_file, delimiter='\t')
        for line in data:
            gafwriter.writerow([item for item in line])


def parse_gaf(filename):
    """Converts a Gene Annotation File (GAF) into a :py:obj:`list` object where every item is a row from the GAF.

    :param file_handle filename: Specify the location of the GAF.
    :return: A list representing the GAF.
    :rtype: :py:obj:`list`
    """
    comment_line = re.compile('^!')
    gaf_array = list()
    with open(os.path.realpath(filename)) as gaf_file:
        for line in csv.reader(gaf_file, delimiter='\t'):
            if not re.match(comment_line, str(line[0])):
                gaf_array.append(line)
    return gaf_array



# !/usr/bin/python3
"""
A subgraph object of an OBOGraph object.
"""
from .dag import OboGraph, AbstractNode
import re


class SubGraph(OboGraph):

    """A subgraph of a provided supergraph with node contents."""

    def __init__(self, super_graph, namespace_filter=None, allowed_relationships=None):
        """`SubGraph initializer. Creates a subgraph object of :class:`gocats.dag.OboGraph`. Leave `namespace_filter`
        and `allowed_relationship` as :py:obj:`None` to create the entire ontology graph. Otherwise, provide filters to
        limit what information is pulled into the subgraph.

        :param obj super_graph: A supergraph object i.e. :class:`gocats.godag.GoGraph`.
        :param str namespace_filter: Specify the namespace of a sub-ontology namespace, if one is available for the ontology.
        :param list allowed_relationships: Specify a list of relationships to utilize in the graph, other relationships will be ignored.
        """
        self.super_graph = super_graph
        if self.super_graph.namespace_filter and self.super_graph.namespace_filter != namespace_filter:
            raise Exception("Unless a namespace_filter is not specified for a parent_graph, a subgraph's namespace_filter must not differ from its parent graph's namespace_filter.\nsubgraph namespace_filter = {}, supergraph namespace_filter = {}").format(namespace_filter, self.super_graph.namespace_filter)
        if self.super_graph.allowed_relationships and allowed_relationships and any(relationship not in self.super_graph.allowed_relationships for relationship in allowed_relationships):
            raise Exception("Unless an allowed_relationships list is not specified for a parent graph, a subgraph's allowed_relationships list must be a subset of, or exactly, its parent graph's allowed_relationships list.\nsubgraph allowed_relationships = {}, supergraph allowed_relationships = {}").format(allowed_relationships, self.super_graph.allowed_relationships)
        super().__init__(namespace_filter, allowed_relationships)
        self.seeded_size = None  # The number of nodes filtered in the keyword search, used for informational purposes only.
        self.category_node = None
        self._root_id_mapping = None
        self._root_node_mapping = None
        self._content_mapping = None
        self._subgraph_finalized = False

    @property
    def root_id_mapping(self):
        """Property describing a mapping :py:obj:`dict` that relates every ontology term ID of subgraphs in
        :class:`gocats.dag.OboGraph` to a :py:obj:`list` of root :class:`gocats.subdag.CategoryNode` IDs.

        :return: :py:obj:`dict` of :class:`gocats.subdag.SubGraphNode` IDs mapped to a :py:obj:`list` of root :class:`gocats.subdag.CategoryNode` IDs.
        :rtype: :py:obj:`dict`
        """
        if (self._modified and self.category_node) or self._root_id_mapping == None:
            self._root_id_mapping = {node.id: self.category_node.id for node in self.category_node.descendants}
            self._root_id_mapping[self.category_node.id] = self.category_node.id
        elif not self.category_node:
            raise Exception("Mapping failed: category node not identified.")
        return self._root_id_mapping

    @property
    def root_node_mapping(self):
        """Property describing a mapping :py:obj:`dict` that relates every ontology :class:`gocats.subdag.SubGraphNode`
        object of subgraphs in :class:`gocats.subdag.SubGraph` to a :py:obj:`list` of root :class:`gocats.subdag.CategoryNode` objects.

        :return: :py:obj:`dict` of :class:`gocats.subdag.SubGraphNode` objects mapped to a :py:obj:`list` of root :class:`gocats.subdag.CategoryNode` objects.
        :rtype: :py:obj:`dict`
        """
        if (self._modified and self.category_node) or self._root_node_mapping == None:
            self._root_node_mapping = {node: self.category_node for node in self.category_node.descendants}
            self._root_node_mapping[self.category_node] = self.category_node
        elif not self.category_node:
            raise Exception("Mapping failed: category node not identified.")
        return self._root_node_mapping

    @property
    def content_mapping(self):
        """Property describing a mapping :py:obj:`dict` that relates every root :class:`gocats.subdag.CategoryNode` IDs
        of subgraphs in a :class:`gocats.subdag.SubGraph` to a :py:obj:`list` of their subgraph nodes' IDs.

        :return: :py:obj:`dict` of :class:`gocats.dag.AbstractNode` IDs mapped to a :py:obj:`list' of :class:`gocats.dag.AbstractNode` IDs.
        :rtype: :py:obj:`dict`
        """
        if (self._modified and self.category_node) or self._content_mapping == None:
            self._content_mapping = {self.category_node.id: [node.id for node in self.category_node.descendants]}
        elif not self.category_node:
            raise Exception("Mapping failed: category node not identified.")
        return self._content_mapping

    def subnode(self, super_node):
        """Defines a :class:`gocats.subdag.SubGraph` node object. Calls :func:`add_node` to convert a supergraph node
        into a :class:`gocats.subdag.SubGraphNode` and add this node to the subgraph.

        :param super_node: A node object from the supergraph i.e. :class:`gocats.godag.GoGraphNode`.
        :return: A :class:`gocats.subdag.SubGraphNode` object.
        :rtype: :py:obj:`class`
        """
        if super_node.id not in self.id_index:
            self.add_node(super_node)
        return self.id_index[super_node.id]

    def add_node(self, super_node):
        """Converts a supergraph node into a :class:`gocats.subdag.SubGraphNode` and adds this node to the subgraph.
        Sets modification state to :py:obj:`True`.

        :param obj super_node: A node object from the supergraph i.e. :class:`gocats.godag.GoGraphNode`.
        :return: None
        :rtype: :py:obj:`None`
        """
        if super_node.id not in self.id_index:
            subgraph_node = SubGraphNode(super_node, self.allowed_relationships)
            if self.valid_node(subgraph_node):
                super().add_node(subgraph_node)
            self._modified = True

    # TODO: Rename/reconsider this (needs to be similar to instantiate_valid_edges)
    def connect_subnodes(self):
        """Analogous to :func:`gocats.dag.instantiate_valid_edges` and :func:`gocats.dag.AbstractEdge.connect_nodes`.
        Updates child and parent node sets for each :class:`gocats.subdag.SubGraphNode` in the
        :class:`gocats.subdag.SubGraph`. Adds edge object references to nodes and node object references to edges.
        Counts instances of relationship IDs and sets modification state to :py:obj:`True`.

        :return: None
        :rtype: :py:obj:`None`
        """
        for subnode in self.node_list:
            subnode.update_children([self.id_index[child.id] for child in subnode.super_node.child_node_set if child.id in self.id_index])
            subnode.update_parents([self.id_index[parent.id] for parent in subnode.super_node.parent_node_set if parent.id in self.id_index])
            for edge in subnode.super_node.edges:  # This counts the number of times each relationship type is used in a subgraph and also adds edges to the subgraph
                if edge.forward_node.id in self.id_index and edge.reverse_node.id in self.id_index:
                    self.add_edge(edge)
                    try:
                        self.relationship_count[edge.relationship.id] += 1
                    except KeyError:
                        self.relationship_count[edge.relationship.id] = 1
        self._modified = True

    def greedily_extend_subgraph(self):
        """Extends a seeded subgraph to include all supergraph descendants of the nodes. Searches through the supergraph
        to add new SubGraphNode objects.

        :return: None
        :rtype: :py:obj:`None`
        """
        possible_graph_extension_nodes = set()
        for node in self.category_node.child_node_set:
            possible_graph_extension_nodes.update(node.super_node.descendants)  # May want to implement this as a generator somehow for efficiency.
        for super_node in possible_graph_extension_nodes:
            if super_node.id not in self.id_index:
                self.add_node(super_node)
        self.connect_subnodes()

    def conservatively_extend_subgraph(self):
        """**Not currently in use.*** Needs to be updated to handle CategoryNode.

        Extends a seeded subgraph to include only nodes in the supergraph that occur along paths between nodes in the
        subgraph. Searches through the supergraph to add new node objects.

        :return: None
        :rtype: :py:obj:`None`
        """
        graph_extension_nodes = set()
        for subleaf in self.leaves:
            start_node = self.super_graph.id_index[subleaf.id]
            end_node = self.super_graph.id_index[self.representative_node.id]
            graph_extension_nodes.update(self.nodes_between(start_node, end_node))
        for super_node in graph_extension_nodes:
            if super_node.id not in self.id_index and self.valid_node(super_node):
                self.add_node(super_node)
        self.connect_subnodes()

    def remove_orphan_paths(self):
        """**Not currently in use.** Needs to be updated ot handle CategoryNode.

        Removes nodes and their descendants from the subgraph which do not root to the
        category-representative node.

        :return: None
        :rtype: :py:obj:`None`
        """
        for orphan in self.orphans:
            orphaned_descendants = orphan.descendants - self.representative_node.descendants
            if orphaned_descendants:
                for descendant in orphaned_descendants:
                    self.remove_node(descendant)
            self.remove_node(orphan)

    @staticmethod
    def find_representative_nodes(subgraph, search_string_list):
        """Compiles a list candidate :class:`gocats.subdag.SubGraphNode` objects from the :class:`gocats.subdag.SubGraph`
        object based on a list of search strings matching strings in the names of the nodes (using regular
        expressions). Returns a list containing a single candidate node with the highest number of descendants when possible, returns the sole
        node if the subgraph only contains one node, returns a list of all seeded nodes when choosing candidates is impossible, or
        aborts if the subgraph is empty.

        :param subgraph: A :class:`gocats.subdag.SubGraph` object.
        :param search_string_list: A :py:obj:`list` of search term :py:obj:`str` entries.
        :return: A list of one or more candidate term :class:`gocats.subgraph.SubGraphNode` chosen as the subgraph's representative ontology term(s).
        """
        representative_nodes = list()
        if len(subgraph.node_list) == 1:  # For the case where there is only one node, seeded. Go ahead and set this to the representative.
            representative_nodes.append(subgraph.node_list[0])
        elif not subgraph.node_list:
            raise Exception("Subgraph did not seed any nodes from the supergraph! Aborting.")
        else:
            candidates = [node for node in subgraph.node_list if any(re.search('(?<!\-)'+search_string+'(?!\-)', node.name) for search_string in search_string_list) and node not in subgraph.leaves and not node.obsolete]
            if candidates:
                representative_node_scoring = {node: len(node.descendants) for node in candidates}
                representative_nodes.append(max(representative_node_scoring, key=representative_node_scoring.get))
            elif not candidates:  # and all(subgraph.node_list) in subgraph.leaves:
                representative_nodes = [node for node in subgraph.node_list]
        return representative_nodes

    @staticmethod
    def from_filtered_graph(super_graph, subgraph_name, keyword_list, namespace_filter=None, allowed_relationships=None, extension='greedy'):
        """Staticmethod for extracting a subgraph from the supergraph by selecting nodes that contain vocabulary in the
        supplied keyword list. Leave `namespace_filter` and `allowed_relationship` as :py:obj:`None` to create the
        entire ontology graph. Otherwise, provide filters to limit what information is pulled into the subgraph.
        Graph `extension` variable defaults to 'greedy' which calls :func:`greedily_extend_subgraph` to add nodes to the
        subgraph after instantiation. Conversely, 'conservative' may be used to call
        :func:`conservatively_extend_subgraph` for this function.

        :param obj super_graph: A supergraph object i.e. :class:`gocats.godag.GoGraph`.
        :param str subgraph_name: The name of the subgraph being created; will be used as the id of the :class:`gocats.subdag.CategoryNode`.
        :param keyword_list: A :py:obj:`list` of :py:obj:`str` entries used to query the supergraph for concepts to be extracted into subgraphs.
        :param str namespace_filter: Specify the namespace of a sub-ontology namespace, if one is available for the ontology.
        :param list allowed_relationships: Specify a list of relationships to utilize in the graph, other relationships will be ignored.
        :param str extension: Specify 'greedy' or 'conservative' to determine how subgraphs will be extended after creation (defaults to greedy).
        :return: A :class:`gocats.subdag.SubGraph` object.
        """
        subgraph = SubGraph(super_graph, namespace_filter, allowed_relationships)
        keyword_list = [word.lower() for word in keyword_list]
        filtered_nodes = super_graph.filter_nodes(keyword_list)
        subgraph.seeded_size = len(filtered_nodes)
        for super_node in filtered_nodes:
            subgraph.add_node(super_node)
        subgraph.connect_subnodes()

        representative_nodes = subgraph.find_representative_nodes(subgraph, keyword_list)
        subgraph.category_node = CategoryNode(subgraph_name, representative_nodes, namespace_filter)
        subgraph.root_nodes.extend(representative_nodes)
        if extension == 'greedy':
            subgraph.greedily_extend_subgraph()
        elif extension == 'conservative':
            subgraph.conservatively_extend_subgraph()
        subgraph_orphans_descendants = set()
        for orphan in subgraph.orphans:
            for node in orphan.descendants:
                subgraph_orphans_descendants.add(node)
        subgraph_orphans_descendants.update([orphan for orphan in subgraph.orphans])
        subgraph._subgraph_finalized = True

        return subgraph


class SubGraphNode(AbstractNode):

    """An instance of a node within a subgraph of an OBO ontology (supergraph)
    """

    def __init__(self, super_node=None, allowed_relationships=None):
        """SubGraphNode initializer. Inherits from :class:`gocats.dag.AbstractNode` and contains a reference to the
        supergraph node it represents e.g. :class:`gocats.godag.GoGraphNode`.

        :param super_node: A node from the `supergraph`.
        :param allowed_relationships: **Not currently used** Used to specify a list of allowable relationships evaluated between nodes.
        """
        self.allowed_relationships = allowed_relationships
        self.super_node = super_node
        self.parent_node_set = set()
        self.child_node_set = set()
        self._modified = True
        self._descendants = None
        self._ancestors = None

    # TODO: add in update_parent_node_set and update_child_node_set with a _modified switch !!!!

    @property
    def super_edges(self):
        """:py:obj:`property` describing the set of edges referenced in the supergraph node, filtered to only those
         edges with nodes in the subgraph node.

        :return: A set of :class:`gocats.subgraph.SubGraphNode` edges that were copied from the supergraph node.
        :rtype: :py:obj:`set`
        """
        edges = set()
        edges.update([edge for edge in self.super_node.edges if edge.parent_node.id in [node.id for node in self.parent_node_set] and edge.child_node.id in [node.id for node in self.child_node_set]])
        return edges

    @property
    def id(self):
        """:py:obj:`property` describing the ID of the supernode

        :return: The ID of a supernode e.g. :class:`gocats.godag.GoGraphNode`
        :rtype: :py:obj:`str`
        """
        return self.super_node.id

    @property
    def name(self):
        """:py:obj:`property` describing the name of the supernode

        :return: The name of a supernode e.g. :class:`gocats.godag.GoGraphNode`
        :rtype: :py:obj:`str`
        """
        return self.super_node.name

    @property
    def definition(self):
        """:py:obj:`property` describing the definition of the supernode

        :return: The definition of a supernode e.g. :class:`gocats.godag.GoGraphNode`
        :rtype: :py:obj:`str`
        """
        return self.super_node.definition

    @property
    def namespace(self):
        """:py:obj:`property` describing the namespace of the supernode

        :return: A namespace of a supernode e.g. :class:`gocats.godag.GoGraphNode`
        :rtype: :py:obj:`str`
        """
        return self.super_node.namespace

    @property
    def obsolete(self):
        """:py:obj:`property` describing whether or not supernode is marked as obsolete.

        :return: :py:obj:`True` or :py:obj:`False`
        """
        return self.super_node.obsolete

    def update_parents(self, parent_set):
        """Updates the parent_node_set with a set of new parents provided. Sets modification state to :py:obj:`True`.

        :param parent_set: A set of parent nodes to be added to this objects parent_node set.
        :return: None
        :rtype: :py:obj:`None`
        """
        self.parent_node_set.update(parent_set)
        self._modified = True

    def update_children(self, child_set):
        """Updates the child_node_set with a set of new children provided. Sets modification state to :py:obj:`True`.

        :param child_set: A set of child nodes to be added to this objects child_node set.
        :return: None
        :rtype: :py:obj:`None`
        """
        self.child_node_set.update(child_set)
        self._modified = True


class CategoryNode(AbstractNode):

    """A special node added to the subgraph which contains all representative nodes identified
    and serves as the single representative of the subgraph which represents a concept.
    """

    def __init__(self, category_name, representative_node_list, namespace_filter=None):
        self.parent_node_set = set()
        self.obsolete = False
        self._modified = True
        self._descendants = None
        self._ancestors = None
        self.namespace_filter = namespace_filter  # Needed to make the node valid.
        if self.namespace_filter:
            self.namesapce = namespace_filter[0]
        self.name = category_name
        self.child_node_set = set(representative_node_list)

    @property
    def id(self):
        if len(self.child_node_set) == 1:
            return next(iter(self.child_node_set)).id
        return self.name




# !/usr/bin/python3
"""
Defines a Gene Ontology-specific graph which may have special properties when compared to other OBO formatted
ontologies.
"""
from .dag import OboGraph, AbstractNode


class GoGraph(OboGraph):

    """A Gene-Ontology-specific graph. GO-specific idiosyncrasies go here."""
    
    def __init__(self, namespace_filter=None, allowed_relationships=None):
        """`GoGraph` initializer. Inherits and specializes properties from :class:`gocats.dag.OboGraph`.

        :param str namespace_filter: Specify the namespace of a sub-ontology namespace, if one is available for the ontology.
        :param list allowed_relationships: Specify a list of relationships to utilize in the graph, other relationships will be ignored.
        """
        self.valid_namespaces = ['cellular_component', 'biological_process', 'molecular_function', None]
        if namespace_filter not in self.valid_namespaces:
            raise Exception("{} is not a valid Gene Ontology namespace.\nPlease select from the following: {}".format(namespace_filter, self.valid_namespaces))

        super().__init__(namespace_filter, allowed_relationships)


class GoGraphNode(AbstractNode):

    """Extends AbstractNode to include GO relevant information."""
    
    def __init__(self):
        """`GoGraphNode` initializer. Inherits all properties from :class:`gocats.dag.AbstractNode`.
        """
        super().__init__()



import gocats
import pytest

this_version = str(gocats.__version__)


def hinderer_categories():
    # use the 01-12-2016 version of go.obo to get same results as manuscript_3.
    gocats.create_subgraphs({
    '<database_file>': '/mlab/data/databases/GeneOntology/01-12-2016/go.obo', '<keyword_file>': './gocats/exampledata/example_subdags.csv',
    '<output_directory>': '/mlab/data/eugene/'+this_version+'_output', '--supergraph_namespace': 'cellular_component',
    '--subgraph_namespace': 'cellular_component', '--supergraph_relationships': None, '--subgraph_relationships': None,
    '--map_supersets': False, '--output_termlist': True, '--go-basic-scoping': None, '--network_table_name': None
    })

hinderer_categories()



#!/usr/bin/env python3

"""
The Gene Ontology Categories Suite (GOcats)
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module provides methods for the creation of directed acyclic concept subgraphs of Gene Ontology, along with methods
for evaluating those subgraphs.

Command line implementation::

    Usage:
        gocats create_subgraphs <database_file> <keyword_file> <output_directory> [--supergraph_namespace=<namespace> --subgraph_namespace=<namespace> --supergraph_relationships=<relationships> --subgraph_relationships=<relationships> --network_table_name=<name> --map_supersets --output_termlist --go_basic_scoping --test]
        gocats categorize_dataset <dataset_file> <term_mapping> <output_directory> <mapped_dataset_filename> [--dataset_type=<GAF> --entity_col=<entity> --go_col=<go> --retain_unmapped_annotations]
        gocats remap_goterms <go_database> <goa_gaf> <ancestor_filename> <namespace_filename> [--allowed_relationships=<relationships> --identifier_column=<column>]
        gocats (-h | --help)
        gocats --version

    Options:
        -h --help                                   Shows this screen.
        --version                                   Shows version.
        <database_file>                             GO term database.
        <keyword_file>                              GO term keywords used for finding related GO terms.
        <output_directory>                          Where everything will be saved.
        --supergraph_namespace=<namespace>          Filters the supergraph to a given namespace.
        --subgraph_namespace=<namespace>            Filters the subgraph to a given namespace.
        --supergraph_relationships=<relationships>  Comma separated relationship types defining which allowed in the supergraph. [default: is_a,part_of,has_part]
        --subgraph_relationships=<relationships>    Comma separated relationship types denote which relationships are allowed in the subgraph. [default: is_a,part_of,has_part]
        --network_table_name=<name>                 Custom name for the output NetworkTable.csv to be used with Cytoscape. [default: NetworkTable.csv]
        --map_supersets                             Maps all terms to all root nodes, regardless of if a root node subsumes another.
        --output_termlist                           Outputs a list of all terms in the supergraph as a JsonPickle file in the output directory.
        --go_basic_scoping                          Creates a GO graph similar to go-basic with only scoping-type relationships (is_a and part_of). WARNING, this supersedes relationship definitions.
        <dataset_file>                              A GO dataset file.
        <term_mapping>                              A dictionary mapping category-defining ontology terms to their subgraph children terms. May be produced by GOcats or another method.
        <mapped_dataset_filename>                   The desired name of the mapped GAF.
        --dataset_type=<filetype>                   Enter file type for dataset [GAF|TSV|CSV]. Defaults to GAF. [default: GAF]
        --entity_col=<entity>                       If CSV or TSV file type, indicate which column the entity IDs are listed. Defaults to 0. [default: 0]
        --go_col=<go>                               If CSV or TSV file type, indicate which column the GO IDs are listed. Defaults to 1. [default: 1]
        --retain_unmapped_annotations               If specified, annotations that are not mapped to a concept are copied into the mapped dataset output file with its original annotation.
        <go_database>                               The Gene Ontology file to use
        <goa_gaf>                                   Gene annotation format file
        <ancestor_filename>                         Where to write the JSONized gene to GO Term relationships to
        <namespace_filename>                        Where to save the JSONized GO Term to namespace relationships to
        --allowed_relationships=<relationships>     Comma separated string of term-to-term relationships to allow [default: is_a,part_of,has_part]
        --identifier_column=<column>                Which column has the gene identifiers [default: 1]
        --test                                      Outputs json files to compare versions of GOcats.
"""

import docopt
from . import gocats
from ._version import __version__


def main(args):
    if args['create_subgraphs']:
        database_file = args['<database_file>']
        keyword_file = args['<keyword_file>']
        output_directory = args['<output_directory>']
        if args['--supergraph_namespace']:
            supergraph_namespace = args['--supergraph_namespace']
        else:
            supergraph_namespace = None
        if args['--subgraph_namespace']:
            subgraph_namespace = args['--subgraph_namespace']
        else:
            subgraph_namespace = None
        if args['--supergraph_relationships']:
            supergraph_relationships = args['--supergraph_relationships'].split(",")
        else:
            supergraph_relationships = ["is_a", "part_of", "has_part"]
        if args['--subgraph_relationships']:
            subgraph_relationships = args['--subgraph_relationships'].split(",")
        else:
            subgraph_relationships = ["is_a", "part_of", "has_part"]
        if args['--map_supersets']:
            map_supersets = True
        else:
            map_supersets = False
        if args['--output_termlist']:
            output_termlist = True
        else:
            output_termlist = False
        if args['--go_basic_scoping']:
            go_basic_scoping = True
        else:
            go_basic_scoping = False
        if args['--test']:
            test = True
        
        gocats.create_subgraphs(database_file, keyword_file, output_directory, supergraph_namespace=None, subgraph_namespace=None, supergraph_relationships=['is_a', 'part_of', 'has_part'], subgraph_relationships=['is_a', 'part_of', 'has_part'], map_supersets=False, output_termlist=False, go_basic_scoping=False, network_table_name=None, test=False)

    elif args['categorize_dataset']:
      
        dataset_file = args['<dataset_file>']
        term_mapping = args['<term_mapping>']
        output_directory = args['<output_directory>']
        mapped_dataset_filename = args['<mapped_dataset_filename>']
        
        if args['--dataset_type']:
            dataset_type = args['--dataset_type']
        else:
            dataset_type = "GAF"
        if args['--entity_col']:
            entity_col = int(args['--entity_col'])
        else:
            entity_col = 0
        if args['--go_col']:
            go_col = int(args['--go_col'])
        else:
            go_col = 1
        if args['--retain_unmapped_annotations']:
            retain_unmapped_annotations = True
        else:
            retain_unmapped_annotations = False
        
        gocats.categorize_dataset(dataset_file, term_mapping, output_directory, mapped_dataset_filename, dataset_type, entity_col, go_col, retain_unmapped_annotations)
        
    elif args['remap_goterms']:
        go_database = args['<go_database>']
        goa_gaf = args['<goa_gaf>']
        ancestor_filename = args['<ancestor_filename>']
        namespace_filename = args['<namespace_filename>']
        if args['--allowed_relationships']:
            allowed_relationships = args['--allowed_relationships'].split(",")
        else:
            allowed_relationships = ["is_a", "part_of", "has_part"]
        if args['--identifier_column']:
            identifier_column = int(args['--identifier_column'])
        else:
            identifier_column = 1

        gocats.remap_goterms(go_database, goa_gaf, ancestor_filename, namespace_filename, allowed_relationships, identifier_column)

if __name__ == '__main__':
    args = docopt.docopt(__doc__, help=True, version=str('GOcats Version ') + __version__)
    main(args)



# !/usr/bin/python3
"""
The Gene Ontology Categories Suite (GOcats)
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module provides methods for the creation of directed acyclic concept subgraphs of Gene Ontology, along with methods
for evaluating those subgraphs.
"""
import os
import sys
import re
import csv
import jsonpickle
import json
from collections import defaultdict
from . import ontologyparser
from . import godag
from . import subdag
from . import tools
from . import _version

jsonpickle.set_encoder_options('json', sort_keys=True, indent=2)
__version__ = _version.__version__


def json_format_graph(graph_object, graph_identifier):
    "Creates a dictionary representing the edges in the graph and formats it in such a way that it can be encoded into JSON for comparing the graph objects between versions of GOcats."
    json_dict = dict()

    for edge in graph_object.edge_list:
        json_dict[str(graph_identifier)+'_'+edge.json_edge[0]] = edge.json_edge[1]

    return json_dict


def build_graph(args):
    """**Not yet implemented**

    Try build_graph_interpreter to create a GO graph object to explore within a Python interpreter."""
    # FIXME: JsonPickle is reaching max recursion depth because of the fact that objects point to one another.
    if args['--supergraph_namespace']:
        supergraph_namespace = args['--supergraph_namespace']
    else:
        supergraph_namespace = None
    if args['--allowed_relationships']:
        allowed_relationships = args['--allowed_relationships']
    else:
        allowed_relationships = None

    database = open(args['<database_file>'], 'r')
    output_directory = args['<output_directory>']
    if not os.path.exists(output_directory):
        os.makedirs(output_directory)
    database_name = os.path.basename(args['<database_file>'])
    graph_class = {'go.obo': godag.GoGraph(supergraph_namespace, allowed_relationships)}
    graph = graph_class[database_name]
    parsing_class = {'go.obo': ontologyparser.GoParser(database, graph)}
    parsing_class[database_name].parse()
    database.close()


def build_graph_interpreter(database_file, supergraph_namespace=None, allowed_relationships=None, relationship_directionality='gocats'):
    """Creates a graph object of GO, which can be traversed and queried within a Python interpreter.

    :param file_handle database_file: Ontology database file.
    :param str supergraph_namespace: Optional - Filter graph to a sub-ontology namespace.
    :param list allowed_relationships: Optional - Filter graph to use only those relationships listed.
    :param relationship_directionality: Optional - Any string other than 'gocats' will retain all original GO relationship directionalities. Defaults to reverseing has_part direction.
    :return: A Graph object of the ontology provided.
    :rtype: :py:obj:`class`
    """
    database = open(database_file, 'r')
    graph = godag.GoGraph(supergraph_namespace, allowed_relationships)
    go_parser = ontologyparser.GoParser(database, graph, relationship_directionality=relationship_directionality)
    go_parser.parse()
    database.close()
    return graph


def create_subgraphs(database_file, keyword_file, output_directory, supergraph_namespace=None, subgraph_namespace=None, supergraph_relationships=['is_a', 'part_of', 'has_part'], subgraph_relationships=['is_a', 'part_of', 'has_part'], map_supersets=False, output_termlist=False, go_basic_scoping=False, network_table_name=None, test=False):
    """Creates a graph object of an ontology, processed into :class:`gocats.dag.OboGraph` or to an object that
    inherits from :class:`gocats.dag.OboGraph`, and then extracts subgraphs which represent concepts that are defined
    by a list of provided keywords. Each subgraph is processed into :class:`gocats.subdag.SubGraph`.

    :param database_file: Ontology database file.
    :param keyword_file: A CSV file with two columns: column 1 naming categories, and column 2 listing search strings (no quotation marks, separated by semicolons).
    :param output_directory: The directory where results are stored.
    :param supergraph_namespace: a supergraph sub-ontology to filter e.g. cellular_component, optional
    :param subgraph_namespace: a subgraph sub-ontology to filter e.g. cellular_component, optional
    :param supergraph_relationships: a list of relationships to limit in the supergraph e.g. ['is_a', 'part_of'], optional
    :param subgraph_relationships: a list of relationships to limit in subgraphs e.g. ['is_a', 'part_of'], optional
    :param map_supersets: whether to allow subgraphs to subsume other subgraphs, logical, optional
    :param output_termlist: whether to create a translation of ontology terms to their names to improve interpretability of dev test results, logical, optional
    :param go-basic-scoping: whether to create a GO graph similar to go-basic with only scoping-type relationships (is_a and part_of), logical, optional
    :param network_table_name: whether to make a specific name for the network table produced from the subgraphs (defaults to NetworkTable.csv)
    :return: None
    :rtype: :py:obj:`None`
    """

    if go_basic_scoping:
        supergraph_relationships = ['is_a', 'part_of']
        subgraph_relationships = ['is_a', 'part_of']

    # Building the supergraph
    database = open(database_file, 'r')
    output_directory = output_directory
    if not os.path.exists(output_directory):
        os.makedirs(output_directory)
    database_name = os.path.basename(database_file)
    graph_class = {'go.obo': godag.GoGraph(supergraph_namespace, supergraph_relationships)}
    try:
        supergraph = graph_class[database_name]
    except KeyError:
        print("The provided ontology filename was not recognized. Please do not rename ontology files. The accepted list of file names are as follows: \n", graph_class.keys())
        sys.exit()
    parsing_class = {'go.obo': ontologyparser.GoParser(database, supergraph)}
    try:
        parsing_class[database_name].parse()
    except KeyError:
        print("The provided ontology filename was not recognized. Please do not rename ontology files. The accepted list of file names are as follows: \n", graph_class.keys())
        sys.exit()
    if output_termlist:
        tools.jsonpickle_save(list(supergraph.id_index.keys()), os.path.join(output_directory, "termlist"))

    id_translation = dict()
    for id, node in supergraph.id_index.items():
        id_translation[id] = node.name
    tools.jsonpickle_save(id_translation, os.path.join(output_directory, "id_translation"))

    database.close()

    # Building and collecting subgraphs
    subgraph_collection = dict()
    with open(keyword_file, newline='') as file:
        reader = csv.reader(file, delimiter=',', quoting=csv.QUOTE_MINIMAL)
        for row in reader:
            subgraph_name = row[0]
            keyword_list = [keyword for keyword in re.split(';', row[1])]
            subgraph_collection[subgraph_name] = subdag.SubGraph.from_filtered_graph(supergraph, subgraph_name, keyword_list, subgraph_namespace, subgraph_relationships)

    # Handling superset mapping
    if not map_supersets:
        category_subsets = find_category_subsets(subgraph_collection)
    else:
        print("NOTE: supersets were mapped.")
        category_subsets = None

    collection_id_mapping = dict()
    collection_node_mapping = dict()
    collection_content_mapping = dict()
    for subgraph_name, subgraph in subgraph_collection.items():
        for node_id, category_node_id in subgraph.root_id_mapping.items():
            try:
                collection_id_mapping[node_id].update([category_node_id])
            except KeyError:
                collection_id_mapping[node_id] = set([category_node_id])
        for node, category_node in subgraph.root_node_mapping.items():
            try:
                collection_node_mapping[node].update(category_node)
            except KeyError:
                collection_node_mapping[node] = {category_node}
        for rep_node, content in subgraph.content_mapping.items():
            collection_content_mapping[rep_node] = content

    # Remove root nodes that are subsets of existing root nodes from mapping
    if category_subsets:
        for node_id, root_id_list in collection_id_mapping.items():
            for subset_id, superset_ids in category_subsets.items():
                if subset_id in root_id_list:
                    [root_id_list.remove(node) for node in superset_ids if node in root_id_list]
    # TODO: do the same for node_object_mapping

    # Save mapping files and create report
    tools.jsonpickle_save(collection_id_mapping, os.path.join(output_directory, "GC_id_mapping"))
    tools.json_save(collection_id_mapping, os.path.join(output_directory, "GC_id_mapping"))
    tools.jsonpickle_save(collection_content_mapping, os.path.join(output_directory, "GC_content_mapping"))
    tools.json_save(collection_content_mapping, os.path.join(output_directory, "GC_content_mapping"))
    with open(os.path.join(output_directory, 'subgraph_report.txt'), 'w') as report_file:
        report_file.write(
            'Subgraph data\nSupergraph filter: {}\nSubgraph filter: {}\nGO terms in the supergraph: {}\nGO terms in subgraphs: {}\nRelationship prevalence: {}'.format(
                supergraph_namespace, subgraph_namespace, len(set(supergraph.node_list)),
                len(set(collection_id_mapping.keys())), supergraph.relationship_count))
        for subgraph_name, subgraph in subgraph_collection.items():
            out_string = """
                -------------------------
                {}
                Subgraph relationships: {}
                Seeded size: {}
                Representative node: {}
                Nodes added: {}
                Non-subgraph hits (orphans): {}
                Total nodes: {}
                """.format(subgraph_name, subgraph.relationship_count, subgraph.seeded_size,
                           [node.name for node in subgraph.category_node.child_node_set], len(subgraph.node_list) - subgraph.seeded_size,
                           len(subgraph.node_list) - len(subgraph.root_id_mapping.keys()),
                           len(subgraph.root_id_mapping.keys()))
            report_file.write(out_string)

    # FIXME: cannot json save due to recursion of objects within objects...
    # tools.jsonpickle_save(collection_node_mapping, os.path.join(args['<output_directory>'], "GC_node_mapping"))

    if network_table_name is None:
        network_table_name = "Network_table.csv"

    # Making a file for network visualization via Cytoscape 3.0
    with open(os.path.join(output_directory, network_table_name), 'w', newline='') as network_table:
        edgewriter = csv.writer(network_table, delimiter=',', quotechar='|', quoting=csv.QUOTE_MINIMAL)
        for term_id, root_id_list in collection_id_mapping.items():
            for root_id in root_id_list:
                edgewriter.writerow([term_id, root_id])

    if test:
        json_graph_destination_file = os.path.join(output_directory, str(__version__)+'_graph_output.json')
        with open(json_graph_destination_file, 'w') as destfile:
            destfile.write(jsonpickle.encode(json_format_graph(supergraph, 'supergraph')))
            for subgraph_name, subgraph in subgraph_collection.items():
                destfile.write(jsonpickle.encode(json_format_graph(subgraph, subgraph_name)))

        jsonpickle_supergraph_destination_file = os.path.join(output_directory, str(__version__)+'_supergraph_output.jsonpickle')
        jsonpickle_clean_graph(supergraph)
        with open(jsonpickle_supergraph_destination_file, 'w') as destfile:
            destfile.write(jsonpickle.encode(supergraph))

        for subgraph_name, subgraph in subgraph_collection.items():
            jsonpickle_clean_graph(subgraph)
            jsonpickle_subgraph_destination_file = os.path.join(output_directory, str(__version__)+'_'+subgraph_name+'_output.jsonpickle')
            with open(jsonpickle_subgraph_destination_file, 'w') as destfile:
                destfile.write(jsonpickle.encode(subgraph))


def jsonpickle_clean_graph(graph):
    # remove circular references from the nodes
    for node in graph.node_list:
        node.edges = None
        if node.parent_node_set:
            node.parent_node_set = sorted([str(node2.id) for node2 in node.parent_node_set])
        if node.child_node_set:
            node.child_node_set = sorted([str(node2.id) for node2 in node.child_node_set])
        if node._descendants:
            node._descendants = sorted([str(node2.id) for node2 in node._descendants])
        if node._ancestors:
            node._ancestors = sorted([str(node2.id) for node2 in node._ancestors])

    # remove sets or unneeded references from the graph
    if hasattr(graph, "super_graph") and graph.super_graph:
        graph.super_graph = None
    if graph._orphans:
        graph._orphans = sorted([str(node2.id) for node2 in graph._orphans])
    if graph._leaves:
        graph._leaves = sorted([str(node2.id) for node2 in graph._leaves])
    for vocab in graph.vocab_index:
        graph.vocab_index[vocab] = sorted([str(node2.id) for node2 in graph.vocab_index[vocab]])
    if graph.used_relationship_set:
        graph.used_relationship_set = sorted([relationship for relationship in graph.used_relationship_set])


# TODO: the workaround for accessing the sole item in the set here is hacky, fix this later.
def find_category_subsets(subgraph_collection):
    """Finds subgraphs which are subsets of other subgraphs to remove redundancy, when specified.

    :param subgraph_collection: A dictionary of subgraph objects (keys: subgraph name, values: subgraph object).
    :return: A dictionary relating which subgraph objects are subsets of other subgraphs (keys: subset subgraph, values: superset subgraphs).
    :rtype: :py:obj:`dict`
    """
    is_subset_of = dict()
    for subgraph in subgraph_collection.values():
        for next_subgraph in subgraph_collection.values():
            if len(subgraph.category_node.child_node_set) == 1 and len(next_subgraph.category_node.child_node_set) == 1:
                if next(iter(subgraph.category_node.child_node_set)).id != next(iter(next_subgraph.category_node.child_node_set)).id and next(iter(subgraph.category_node.child_node_set)).id in next_subgraph.root_id_mapping.keys():
                    try:
                        is_subset_of[next(iter(subgraph.category_node.child_node_set)).id].add(next(iter(next_subgraph.category_node.child_node_set)).id)
                    except KeyError:
                        is_subset_of[next(iter(subgraph.category_node.child_node_set)).id] = {next(iter(next_subgraph.category_node.child_node_set)).id}
    return is_subset_of


def categorize_dataset(dataset_file, term_mapping, output_directory, mapped_dataset_filename, dataset_type="GAF", entity_col=0, go_col=1, retain_unmapped_annotations=False):
    """Reads in a Gene Annotation File (GAF) and maps the annotations contained therein to the categories organized by
    GOcats or other methods. Outputs a mapped GAF and a list of unmapped genes in the specified output directory.

    :param dataset_file: A file containing gene annotations.
    :param term_mapping: A dictionary mapping category-defining ontology terms to their subgraph children terms. May be produced by GOcats or another method.
    :param output_directory: The directory where the output file will be stored.
    :param mapped_dataset_filename: The desired name of the mapped GAF.
    :param dataset_type: Enter file type for dataset [GAF|TSV|CSV]. Defaults to "GAF".
    :param entity_col: If CSV or TSV file type, indicate which column the entity IDs are listed. Defaults to 0.
    :param go_col: If CSV or TSV file type, indicate which column the GO IDs are listed. Defaults to 1.
    :param retain_unmapped_annotations: If specified, annotations that are not mapped to a concept are copied into the mapped dataset output file with its original annotation.
    :return: None
    :rtype: :py:obj:`None`
    """
    

    mapping_dict = tools.jsonpickle_load(term_mapping)
    output_directory = os.path.realpath(output_directory)
    unmapped_entities = set()
    if not os.path.exists(output_directory):
        os.makedirs(output_directory)

    if dataset_type == "GAF" or not dataset_type:
        loaded_gaf_array = tools.parse_gaf(dataset_file)
        mapped_gaf_array = list()
        for line in loaded_gaf_array:
            if line[4] in mapping_dict.keys():
                mapped_terms = mapping_dict[line[4]]
                for term in mapped_terms:
                    mapped_gaf_array.append(line[0:4] + [term] + line[5:-1])
            else:
                if line[2] == '':
                    unmapped_entities.add('NO_GENE:' + line[1])
                else:
                    unmapped_entities.add(line[2])
        tools.write_out_gaf(mapped_gaf_array, os.path.join(output_directory, mapped_dataset_filename))
        tools.list_to_file(os.path.join(output_directory, mapped_dataset_filename + '_unmappedEntities'), unmapped_entities)

    elif dataset_type == "CSV":
        mapped_rows = []
        with open(os.path.realpath(dataset_file), 'r') as csv_file:
            csv_reader = csv.reader(csv_file, delimiter=',')
            header = next(csv_reader)
            for row in csv_reader:
                mapped_row = row
                if mapped_row[go_id_index] in mapping_dict.keys():
                    for concept_term in mapping_dict[mapped_row[go_id_index]]:
                        mapped_row[go_id_index] = concept_term
                        mapped_rows.append(mapped_row)
                else:
                    unmapped_entities.add(mapped_row[entity_id_index])
                    if retain_unmapped_annotations:
                        mapped_rows.append(mapped_row)
            mapped_rows.insert(0, header)
        with open(os.path.join(output_directory, mapped_dataset_filename), 'w') as output_csv:
            csv_writer = csv.writer(output_csv, delimiter=',')
            for row in mapped_rows:
                csv_writer.writerow([item for item in row])
        tools.list_to_file(os.path.join(output_directory, 'unmappedEntities'), unmapped_entities)


def remap_goterms(go_database, goa_gaf, ancestor_filename, namespace_filename, allowed_relationships, identifier_column):
    """Reads in a Gene Ontology relationship file, and a Gene Annotation File (GAF), and
    follows the GOcats rules for allowed term-to-term relationships. Generates as output
    a new GAF, and a new term to ontology namespace mapping.
    
    :param go_database: the gene ontology dataset
    :param goa_gaf: the gene annotation file
    :param ancestor_filename: the output file containing new gene to ontology mappings
    :param namespace_filename: the output file containing the term to ontology mappings
    :param allowed_relationships: what term to term relationships will be considered (is_a,part_of,has_part) 
    :param identifier_column: which column is being used for the gene identifiers (1)
    :return: None
    :rtype: :py:obj:`None`
    """
    
    graph = build_graph_interpreter(go_database, allowed_relationships=allowed_relationships)
    gaf_array = tools.parse_gaf(goa_gaf)
    goa_gene_annotation_dict = defaultdict(set)
    # Building the annotation dictionary
    for line in gaf_array:
        goa_gene_annotation_dict[line[identifier_column]].add(line[4]) # the dictionary has DB object symbol  keys and a set of go terms as values
    ancestor_dict = defaultdict(set)
    missing_go_terms = set()
    # Adding ancestors to the annotation dictionary.
    for gene_symbol, go_term_set in goa_gene_annotation_dict.items():
        ancestor_dict[gene_symbol].update(go_term_set)
        for go_term in go_term_set:
            if go_term in graph.id_index.keys():
                ancestor_dict[gene_symbol].update([node.id for node in graph.id_index[go_term].ancestors])
            else:
                missing_go_terms.add(go_term)  # NOTE: These are missing because they are depreciated IDs that are now ALT IDs of another term. Need to incorporate alt ids in GOcats.
    # Need to convert dict sets into lists for json
    ancestor_dict = {gene_symbol: list(go_term_set) for gene_symbol, go_term_set in ancestor_dict.items()}
    # Writeout output json file.
    with open(ancestor_filename, "w") as output_file:
        json.dump(ancestor_dict, output_file)
    with open(namespace_filename, "w") as output_file:
        namespace_translation = {}
        for node in graph.node_list:
            namespace_translation[node.id] = node.namespace
        json.dump(namespace_translation, output_file)
    return None



# !/usr/bin/python3
"""
A parser which reads ontologies in the OBO format and calls appropriate graph objects to store information in a graph
representation. Separate parsing classes within this module operate on distinct ontologies in the OBO Foundry to handle
any subtle differences among ontologies.
"""
import re
from .dag import AbstractEdge, DirectionalRelationship
from .godag import GoGraphNode


class OboParser(object):

    """A scaffolding for parsing OBO formatted ontologies. Contains regular expressions for the basic stanzas and
    information pertinent for creating a graph object of an ontology."""

    def __init__(self):
        """`OboParser` initializer. Contains Regular Expressions for identifying crucial information from OBO formatted
        ontologies.
        """
        self.term_stanza = re.compile('\[Term\]')
        self.go_term = re.compile('GO\:\d{7}')
        self.stanza_id = re.compile('^id:')
        self.stanza_name = re.compile('^name:')
        self.namespace = re.compile('^namespace:')
        self.term_definition = re.compile('^def:')
        self.obsolete = re.compile('^is_obsolete:\strue')
        self.is_a = re.compile('^is_a:')
        self.relationship_match = re.compile('^relationship:')
        self.end_stanza = re.compile('^\s+')
        self.typedef_stanza = re.compile('^\[Typedef\]')
        self.inverse_tag = re.compile('^inverse_of:')

        # May use later.
        # self.comment = re.compile('\!.+')
        # self.subset = re.compile('^subset:')


class GoParser(OboParser):

    """An ontology parser specific to Gene Ontology"""


    def __init__(self, database_file, go_graph, relationship_directionality='gocats'):
        """`GoParser` initializer. Parses a Gene Ontology database file and adds properties found therein to a
        :class:`godag.GoGraph` object. **Importantly:** includes descriptions of semantic directionality of all GO
        relationships.
        :param file_handle database_file: Specify the location of a Gene Ontology .obo file.
        :param go_graph: :class:`gocats.godag.GoGraph` object.
        :return: None
        :rtype: :py:obj:`None`
        """
        super().__init__()
        self.database_file = database_file
        self.go_graph = go_graph
        # 5 types of relationships: scoping, scaling (grouped with scoping for now), spatiotemporal, active, equivalence (not present here), and other.
        # 1 means that the relationship directionality is conventional, 0 means that the semantic directionality points from node 2 to node 1.
        if relationship_directionality == 'gocats':
            self.relationship_mapping = {"ends_during": ("spatiotemporal", 1), "happens_during": ("spatiotemporal", 1), "has_part": ("scoping", 0),
                                         "negatively_regulates": ("active", 1), "never_in_taxon": ("other", 1), "occurs_in": ("spatiotemporal", 1),
                                         "part_of": ("scoping", 1), "positively_regulates": ("active", 1), "regulates": ("active", 1),
                                         "starts_during": ("spatiotemporal", 1), "is_a": ("scoping", 1), "term_tracker_item": ("other", 1)}
        else:
            self.relationship_mapping = {"ends_during": ("spatiotemporal", 1), "happens_during": ("spatiotemporal", 1), "has_part": ("scoping", 1),
                                         "negatively_regulates": ("active", 1), "never_in_taxon": ("other", 1), "occurs_in": ("spatiotemporal", 1),
                                         "part_of": ("scoping", 1), "positively_regulates": ("active", 1), "regulates": ("active", 1),
                                         "starts_during": ("spatiotemporal", 1), "is_a": ("scoping", 1), "term_tracker_item": ("other", 1)}

    def parse(self):
        """Parses the ontology database file and accesses the ontology graph object to add information found in the
        database. Once all information is added, this function calls the graph's instantiate_valid_edges function to
        connect all nodes in the graph by their edges.

        :return: None
        :rtype: :py:obj:`None`
        """
        is_term = False
        is_typedef = False

        for line in self.database_file:

            if not is_term and not is_typedef and re.match(self.term_stanza, line):
                is_term = True
                node = GoGraphNode()
                node_edge_list = []

            if not is_typedef and not is_term and re.match(self.typedef_stanza, line):
                is_typedef = True
                relationship_obj = DirectionalRelationship()

            elif is_term:
                if re.match(self.stanza_id, line):
                    node.id = re.findall(self.go_term, line)[0]
                    curr_stanza_id = node.id

                elif re.match(self.stanza_name, line):
                    node.name = line[6:-1].lower()  # ignores 'name: '

                elif re.match(self.namespace, line):
                    node.namespace = line[11:-1].lower()  # ignores 'namespace: '

                elif re.match(self.term_definition, line):
                    node.definition = re.findall('\"(.*?)\"', line)[0].lower()  # This pattern matches the definition listed within quotes on the line.

                elif re.match(self.is_a, line):
                    node_edge = AbstractEdge(curr_stanza_id, re.findall(self.go_term, line)[0], 'is_a')  # node1, node2, relationship
                    node_edge_list.append(node_edge)
                    if "is_a" not in self.go_graph.relationship_index:
                        is_a_relationship = DirectionalRelationship()
                        is_a_relationship.id = "is_a"
                        is_a_relationship.name = "is a"
                        is_a_relationship.category = self.relationship_mapping["is_a"][0]
                        is_a_relationship.direction = self.relationship_mapping["is_a"][1]
                        self.go_graph.used_relationship_set.add(is_a_relationship.id)
                        self.go_graph.add_relationship(is_a_relationship)

                elif re.match(self.relationship_match, line):
                    relationship_id = re.findall("[\w]+", line)[1]  # line example: relationship: part_of GO:0040025 ! vuval development
                    node_edge = AbstractEdge(curr_stanza_id, re.findall(self.go_term, line)[0], relationship_id)
                    node_edge_list.append(node_edge)

                elif re.match(self.obsolete, line):
                    node.obsolete = True

                elif re.match(self.end_stanza, line):
                    if self.go_graph.valid_node(node):
                        self.go_graph.add_node(node)
                        for edge in node_edge_list:
                            if not self.go_graph.allowed_relationships or edge.relationship_id in self.go_graph.allowed_relationships:
                                self.go_graph.add_edge(edge)
                                self.go_graph.used_relationship_set.add(edge.relationship_id)
                        if node_edge_list == [] and node.obsolete == False:  # Have to look at the local edge list because nodes have not been linked with edges yet. Entire graph must be populated first. This is the only way to do this on-the-fly.
                            self.go_graph.root_nodes.append(node)  # make root nodes a set of all namespaces used in the ontology.
                    is_term = False

            elif is_typedef:
                if re.match(self.stanza_id, line):
                    relationship_obj.id = re.findall(r"[\w+\:]+", line)[1]

                elif re.match(self.stanza_name, line):
                    relationship_obj.name = re.findall(r"[\w+\:]+", line)[1]

                elif re.match(self.inverse_tag, line):
                    relationship_obj.inverse_relationship_id = re.findall(r"[\w+\:]+", line)[1]

                elif re.match(self.end_stanza, line):
                    properties = self.relationship_mapping[relationship_obj.id]
                    relationship_obj.category = properties[0]
                    relationship_obj.direction = properties[1]
                    self.go_graph.add_relationship(relationship_obj)
                    is_typedef = False

        self.go_graph.instantiate_valid_edges()
